;; Generated package description from phps-mode.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "phps-mode" "0.4.52.0.20260817.2" "Major mode for PHP with code intelligence" '((emacs "26")) :commit "d509c7228a145fe3eabd40723f59c07fa81597d6" :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :keywords '("tools" "convenience") :url "https://forgejo.cvj.se/cjohansson/emacs-phps-mode")
